package HLstats_Server;
#
# HLstatsX - Real-time player and clan rankings and statistics for Half-Life:Source
# http://www.hlstatsx.com/
# Copyright (C) 2005 Tobias Oetzel (Tobi@hlstatsx.com)
#
# HlstatsX is an enhanced version of HLstats made by Simon Garner
# HLstats - Real-time player and clan rankings and statistics for Half-Life
# http://sourceforge.net/projects/hlstats/
# Copyright (C) 2001  Simon Garner
#             
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
use POSIX;
use IO::Socket;

sub new
{
	my ($class_name, $serverId, $address, $port, $server_name, $rcon_pass, $game) = @_;
	
	my ($self) = {};
	
	bless($self, $class_name);
	
	$self->{id}             = $serverId;
	$self->{address}        = $address;
	$self->{port}           = $port;
	$self->{game}           = $game;
	$self->{rcon}           = $rcon_pass;
    $self->{rcon_obj}	    = null;
    $self->{name}           = $server_name;
    $self->{auto_ban}       = 0;
    $self->{contact}        = "";
    $self->{hlstats_url}    = "";
    
    $self->{last_heartbeat} = 0;
    
	$self->{lines}         = 0;
	$self->{map}           = "";
	$self->{numplayers}    = 0;
	$self->{minplayers}    = 6;
	$self->{maxplayers}    = 0;
	
    $self->{players}       = 0;
 	$self->{rounds}        = 0;
	$self->{kills}         = 0;
	$self->{suicides}      = 0;
	$self->{headshots}     = 0;
	$self->{ct_shots}      = 0;
	$self->{ct_hits}       = 0;
	$self->{ts_shots}      = 0;
	$self->{ts_hits}       = 0;
	$self->{bombs_planted} = 0;
	$self->{bombs_defused} = 0;
	$self->{ct_wins}       = 0;
	$self->{ts_wins}       = 0; 
	$self->{map_started}   = time();
	$self->{map_changes}   = 0;
	$self->{map_rounds}    = 0;
	$self->{map_ct_wins}   = 0;
	$self->{map_ts_wins}   = 0; 
	$self->{map_ct_shots}  = 0;
	$self->{map_ct_hits}   = 0;
	$self->{map_ts_shots}  = 0; 
	$self->{map_ts_hits}   = 0; 

	# team balancer
	$self->{ba_enabled}       = 0;
	$self->{ba_ct_wins}       = 0;
	$self->{ba_ts_win}        = 0;
	$self->{ba_ct_frags}      = 0;
	$self->{ba_ts_frags}      = 0;
	$self->{ba_winner}        = ();
	$self->{ba_map_rounds}    = 0;
	$self->{ba_last_swap}     = 0;
	$self->{ba_player_switch} = 0;  # player switched on his own
	
	$self->{rawsocket_support}            = 0;
	$self->{rawsocket_helpnotice}         = 0;
	$self->{show_stats}                   = 0;
	$self->{broadcasting_events}          = 0;
	$self->{broadcasting_player_actions}  = 0;
	$self->{broadcasting_command}         = "";
 	$self->{broadcasting_command_steamid} = 0;
	$self->{player_events}                = 0; 
 	$self->{player_command}               = 0;
 	$self->{player_command_steamid}       = 0;
	$self->{player_admin_command}         = 0;
	$self->{broadcasting_serverdata}      = 0;
	$self->{broadcasting_server_address}  = "";
	$self->{broadcasting_server_port}     = 0;
	$self->{broadcasting_server_interval} = 100;
	$self->{broadcasting_server_resolved} = "";
	
	$self->{total_kills}                  = 0;
	$self->{total_headshots}              = 0;
	$self->{total_suicides}               = 0;
	$self->{total_rounds}                 = 0;
	$self->{total_shots}                  = 0;
	$self->{total_hits}                   = 0;
	
	$self->{track_server_load}            = 0;
	$self->{track_server_timestamp}       = 0;
	
	$self->{ignore_nextban}               = 0;
	$self->{use_browser}                  = 0;
	
	$self->{round_status}                 = 0;
	$self->{min_players_rank}             = 1;
	
	$self->{admins}                       = ();
	$self->{ignore_bots}                  = 0;
	$self->{tk_penalty}                   = 0;
	$self->{suicide_penalty}              = 0;
	
	
	$self->updateDB();
	
	if ($self->get("rcon"))
	{
	  $self->init_rcon();
#	  $self->{rcon_obj}->execute("cvarlist", 1);
	} 
	
	return $self;
}

sub is_admin
{
  my($self, $steam_id) = @_;
  for (@{$self->{admins}}) { 
    if ($_ eq $steam_id) {
      return 1;
    }   
  }  
  return 0;
}


#
# Set property 'key' to 'value'
#

sub set
{
	my ($self, $key, $value) = @_;
	
	if (defined($self->{$key}))
	{
		if ($self->get($key) eq $value)
		{
			if ($g_debug > 2)
			{
				&::printNotice("Hlstats_Server->set ignored: Value of \"$key\" is already \"$value\"");
			}
			return 0;
		}
		
		$self->{$key} = $value;
		
		if ($key eq "hlstats_url")  {
        	# so ingame browsing will work correctly
            $self->{ingame_url}  = $value;
            $self->{ingame_url}  =~ s/\/hlstats.php//;
         	&::printEvent("SERVER", "Ingame-URL: ".$self->{ingame_url}, 1);
		}
		return 1;
	}
	else
	{
		warn("HLstats_Server->set: \"$key\" is not a valid property name\n");
		return 0;
	}
}


#
# Increment (or decrement) the value of 'key' by 'amount' (or 1 by default)
#

sub increment
{
	my ($self, $key, $amount) = @_;
	
	$amount = int($amount);
	$amount = 1 if ($amount == 0);
	
	my $value = $self->get($key);
	$self->set($key, $value + $amount);
}


#
# Get value of property 'key'
#

sub get
{
	my ($self, $key) = @_;
	
	if (defined($self->{$key}))
	{
		return $self->{$key};
	}
	else
	{
		warn("HLstats_Server->get: \"$key\" is not a valid property name\n");
	}
}


sub init_rcon
{
	my ($self)      = @_;
    my $server_ip   = $self->get("address");
    my $server_port = $self->get("port");
    my $rcon_pass   = $self->get("rcon");
    my $game        = $self->get("game");
	if ($rcon_pass)
	{
      $self->{rcon_obj} = new TRcon($self);
	}  
   	if ($self->{rcon_obj}) 
	{
      &::printEvent ("RCON", "Connecting to rcon on $server_ip:$server_port ... ok");
	  &::printEvent("SERVER", "Server running map: ".$self->get_map(), 1);
	}
}

sub dorcon
{
	my ($self, $command)      = @_;
    my $result;
    my $rcon_obj = $self->get("rcon_obj");
	if ($rcon_obj)  {
      &::printNotice("RCON", $command, 1);
      $result = $rcon_obj->execute($command);
    } else {
      &::printNotice("Rcon error: No Object available");
    }
    return $result;
}	

sub rcon_getaddress
{
	my ($self, $uniqueid) = @_;
    my $result;
    my $rcon_obj = $self->get("rcon_obj");
	if ($rcon_obj)  
	{
	  $result = $rcon_obj->getPlayer($uniqueid);
    } else {
      &::printNotice("Rcon error: No Object available");
    }
    return $result;
}

sub rcon_getStatus
{
	my ($self) = @_;
    my $rcon_obj = $self->get("rcon_obj");
    my $map_result;
    my $max_player_result = -1;
	if ($rcon_obj)  
	{
	  ($map_result, $max_player_result) = $rcon_obj->getServerData();
    } else {
      &::printNotice("Rcon error: No Object available");
    }
    return ($map_result, $max_player_result);
}

sub rcon_getplayers
{
	my ($self) = @_;
    my %result;
    my $rcon_obj = $self->get("rcon_obj");
	if ($rcon_obj)  
	{
	  %result = $rcon_obj->getPlayers();
    } else {
      &::printNotice("Rcon error: No Object available");
    }
    return %result;
}

sub track_server_load
{
  my ($self) = @_;

  if ($self->get("track_server_load") > 0)
  {
    my $last_timestamp = $self->get("track_server_timestamp");
    my $new_timestamp  = time();
    if ($last_timestamp > 0)
    {
      if ($last_timestamp+300 < $new_timestamp)
      {
        my $server_id    = $self->get("id");
        my $act_players  = $self->get("numplayers");
        my $min_players  = $self->get("minplayers");  
        my $max_players  = $self->get("maxplayers");
   	    &::doQuery("INSERT IGNORE INTO hlstats_server_load SET server_id='".$server_id."', timestamp='".$new_timestamp."', act_players='".$act_players."', min_players='".$min_players."', max_players='".$max_players."'");
	    $self->set("track_server_timestamp", $new_timestamp);
        &::printEvent("SERVER", "Insert new server load timestamp", 1);
	  }   
	} else {
      $self->set("track_server_timestamp", $new_timestamp);
	}  
  
  }
   

}

sub dostats 
{
	my ($self) = @_;
    my $rcon_obj = $self->get("rcon_obj");

	if ($rcon_obj)  
	{
      if ($self->get("broadcasting_events") == 1)
      {
        my $hpk = sprintf("%.0f", 0);
        if ($self->{total_kills} > 0) {
          $hpk = sprintf("%.2f", (100/$self->{total_kills})*$self->{total_headshots});
        }  
   	    if ($self->get("broadcasting_command_steamid") == 1) {
          foreach $player (values(::g_players))	{
            if (($player->{server} eq $self->{address}.":".$self->{port}) && ($player->{is_bot} == 0))  {
               my $p_steamid  = $player->get("plain_uniqueid");
               $self->dorcon($self->get("broadcasting_command")." \"$p_steamid\" Tracking ".$self->{players}." players with ".$self->{total_kills}." kills and ".$self->{total_headshots}." headshots ($hpk%)");
            }
          }  
        } else {
 		  $self->dorcon($self->get("broadcasting_command")." Tracking ".$self->{players}." players with ".$self->{total_kills}." kills and ".$self->{total_headshots}." headshots ($hpk%)");
	  	}  
      }  
    }  
}	  

sub get_map
{
  my ($self) = @_;
  if (($self->{map} eq "") || ($self->{maxplayers} == 0)) {
    &::printNotice("get_server status (map, maxplayers)");
    ($self->{map}, $self->{maxplayers}) = $self->rcon_getStatus();
    $self->update_max_players();
    &::printNotice("get_server_status successfully");
    $self->updateDB();
  }
  return $self->{map};
}

sub update_max_players  
{
  my ($self) = @_;
  &::printEvent("SERVER", "Update Max Players", 1);
  my $serverid   = $self->get("id");
  my $maxplayers = $self->get("maxplayers");
  if ($maxplayers > 0) {
    my $query = "
	  UPDATE
		 hlstats_Servers
  	  SET
	     max_players=$maxplayers 
	  WHERE
	     serverId='$serverid'
    ";
    &::doQuery($query);
  }  
  
}

sub update_players
{
  my ($self) = @_;

  &::printNotice("update_players");
  &::printEvent("RCON", "Update Players", 1);
  my %players = $self->rcon_getplayers();
  while ( my($pl, $player) = each(::g_players) )
  {	 
    my $pl_id = $player->get("plain_uniqueid");
    if (defined($players{$pl_id})) 
    {
      my $ping = $players{$pl_id}->{"Ping"};
      $player->set("ping", $ping);
    }
  }
}  


sub update_players_pings
{
  my ($self) = @_;

  if ($self->{numplayers} < $self->{minplayers}) 
  {
    &::printNotice("(IGNORED) NOTMINPLAYERS: Update_player_pings");
  } 
  else 
  {
    &::printNotice("update_player_pings");
    &::printEvent("RCON", "Update Player pings", 1);
    my %players = $self->rcon_getplayers();
    while ( my($pl, $player) = each(::g_players) )
    {	 
       my $pl_id = $player->get("plain_uniqueid");
       if (defined($players{$pl_id})) 
       {
         my $ping = $players{$pl_id}->{"Ping"};
         $player->set("ping", $ping);
         &::recordEvent(
                        "Latency", 0,
                        $player->get("playerid"),
                        $ping
                       );
       }
    }
    &::printNotice("update_player_pings successfully");
  }  
}

sub send_heartbeat
{
	my ($self)  = @_;
	my $q_version = "1.00";

   	if ($self->get("broadcasting_serverdata") > 0)
   	{
	    my $host = $self->get("broadcasting_server_address");
		if ($self->get("broadcasting_server_resolved") eq "") {
	  	  if ($host =~ /[.a-zA-Z]/) { 
	        $host_data = (gethostbyname($host))[4];
	        $self->set("broadcasting_server_resolved", join(".", unpack("C4", $host_data)));
	      } else {
	        $self->set("broadcasting_server_resolved", $host);
	      }  
	    }  
	    my $address        = $self->{broadcasting_server_resolved};
		my $port           = $self->get("broadcasting_server_port");
	    my $dest           = sockaddr_in($port, inet_aton($address));
	    
		my $s_address        = $self->get("address");
		my $s_port           = $self->get("port");
		my $s_name           = $self->get("name");
		my $s_contact        = $self->get("contact");
		my $s_url            = $self->get("hlstats_url");
		my $s_version        = $::g_version;
		my $s_id             = $self->get("id");
		my $s_auto_ban       = $self->get("auto_ban");
		my $s_team_bal       = $self->get("ba_enabled");
		
		my $s_act_players = $self->get("numplayers");
		my $s_max_players = $self->get("maxplayers");
		my $s_min_players = $self->get("minplayers");
		my $s_game        = $self->get("game");
		my $s_rounds      = $self->get("total_rounds");
		my $s_suicides    = $self->get("total_suicides");
		my $s_map         = $self->get("map");
		my $s_map_time    = $self->get("map_started");
		my $s_players     = $self->get("players");
		my $s_kills       = $self->get("total_kills");
		my $s_headshots   = $self->get("total_headshots");
		my $s_shots       = $self->get("total_shots");
		my $s_hits        = $self->get("total_hits");
		my $s_min_rank    = $self->get("min_players_rank");
		
		
	    
	    $msg = $s_name.chr(0xff).$s_address.chr(0xff).$s_port.chr(0xff).$s_contact.chr(0xff).$s_url.chr(0xff).$s_version.chr(0xff).
	           $s_id.chr(0xff).$s_act_players.chr(0xff).$s_max_players.chr(0xff).$s_min_players.chr(0xff).$s_auto_ban.chr(0xff).
	           $s_team_bal.chr(0xFF).$s_game.chr(0xff).$s_min_rank.chr(0xff).$s_rounds.chr(0xff).$s_suicides.chr(0xff).$s_map.chr(0xff).$s_map_time.chr(0xff).
	           $s_players.chr(0xff).$s_kills.chr(0xff).$s_headshots.chr(0xff).$s_shots.chr(0xff).$s_hits.chr(0xff);
	    $msg = chr(0xff).chr(0xff)."S".chr(0xff).$q_version.chr(0xff).$msg;
	   	$bytes = send($::s_socket, $msg, 0, $dest);
	   	&::printEvent("MASTER", "Sent $bytes bytes to master server at '$address:$port' [Serverdata]", 1);
	   	
	   	
	   	if ($self->get("broadcasting_serverdata") > 2)
	   	{
	      $player_msg       = "";
	      $i = 0;
	      while (my($pl, $player) = each(::g_players) )
	      {
	        if ($self->get("id") == $player->{server_id}) 
	        {
			  my $p_id            = $player->get("playerid");        
			  my $p_steamid       = $player->get("plain_uniqueid");        
		      my $p_name          = $player->get("name");
	          my $p_team          = $player->get("team");
	          my $p_kills         = $player->get("map_kills");
			  my $p_deaths        = $player->get("map_deaths");
			  my $p_suicides      = $player->get("map_suicides");
			  my $p_headshots     = $player->get("map_headshots");
			  my $p_shots         = $player->get("map_shots");
			  my $p_hits          = $player->get("map_hits");
	          my $p_ping          = $player->get("ping");
	 	      my $p_connected     = $player->get("connect_time");
		      my $p_skill_change  = $player->get("session_skill");
		      my $p_skill         = $player->get("skill");
	          $player_msg .= $p_id.chr(0xfe).$p_steamid.chr(0xfe).$p_name.chr(0xfe).$p_team.chr(0xfe).$p_kills.chr(0xfe).
	                         $p_deaths.chr(0xfe).$p_suicides.chr(0xfe).$p_headshots.chr(0xfe).$p_shots.chr(0xfe).
	                         $p_hits.chr(0xfe).$p_ping.chr(0xfe).$p_connected.chr(0xfe).$p_skill_change.chr(0xfe).$p_skill.chr(0xff);
	          $i++;               
	          if ($i % 10 == 0)
	          {
			    $player_msg = chr(0xff).chr(0xff)."P".chr(0xff).$q_version.chr(0xff).$s_address.chr(0xff).$s_port.chr(0xff).$player_msg;
	      	    $bytes = send($::s_socket, $player_msg, 0, $dest);
		        &::printEvent("MASTER", "Sent $bytes bytes to master server at '$address:$port' [Playerdata]", 1);
		        $player_msg = "";
	          }
	        }
	      }  
	   	  if ($player_msg ne "")
	   	  {
	        $player_msg = chr(0xff).chr(0xff)."P".chr(0xff).$q_version.chr(0xff).$s_address.chr(0xff).$s_port.chr(0xff).$player_msg;
	 	    $bytes = send($::s_socket, $player_msg, 0, $dest);
		    &::printEvent("MASTER", "Sent $bytes bytes to master server at '$address:$port' [Playerdata]", 1);
	  	  }  
	   	}
	
	   	if ($self->get("broadcasting_serverdata") > 4) # send top-players
	   	{
	      $player_msg = "";
	      $i = 0;
	      
	       my $result = &::doQuery("
	          SELECT
	                 hlstats_Players.playerId, hlstats_Players.skill, hlstats_Players.lastName, hlstats_Players.kills, hlstats_Players.deaths, 
	                 hlstats_Players.suicides, hlstats_Players.headshots, hlstats_Players.hits, 
	                 hlstats_Players.shots, hlstats_PlayerUniqueIds.uniqueId as steamid
	          FROM
	                 hlstats_Players,
	                 hlstats_PlayerUniqueIds
	          WHERE
	                 hlstats_Players.game = '".$s_game."' AND
	                 hlstats_Players.playerId = hlstats_PlayerUniqueIds.playerId AND
	                 hlstats_Players.hideranking = 0        
	          ORDER BY
	                 hlstats_Players.skill DESC
	          LIMIT 0, 100
	       ");
	       while (my($player_id, $skill, $lastName, $kills, $deaths, $suicides, $headshots, $hits, $shots, $steam_id) = $result->fetchrow_array)
	  	   {
	          $player_msg .= $player_id.chr(0xfe).$steam_id.chr(0xfe).$lastName.chr(0xfe).$kills.chr(0xfe).$deaths.chr(0xfe).$suicides.chr(0xfe).$headshots.chr(0xfe).
	                         $shots.chr(0xfe).$hits.chr(0xfe).$skill.chr(0xff);
	          $i++;               
	          if ($i % 10 == 0)
	          {
			    $player_msg = chr(0xff).chr(0xff)."T".chr(0xff).$q_version.chr(0xff).$s_game.chr(0xff).$player_msg;
	      	    $bytes = send($::s_socket, $player_msg, 0, $dest);
		        &::printEvent("MASTER", "Sent $bytes bytes to master server at '$address:$port' [Top-Players]", 1);
		        $player_msg = "";
	          }
	  	   }
	   	  if ($player_msg ne "")
	   	  {
	        $player_msg = chr(0xff).chr(0xff)."T".chr(0xff).$q_version.chr(0xff).$s_game.chr(0xff).$player_msg;
	 	    $bytes = send($::s_socket, $player_msg, 0, $dest);
		    &::printEvent("MASTER", "Sent $bytes bytes to master server at '$address:$port' [Top-Players]", 1);
	  	  }  
	
	   	}
	   	$self->set("last_heartbeat", time());
   	}
   	
}


sub send_udp_data
{
	my ($self, $msg, $address, $port) = @_;
	
	if ($self->get("rawsocket_support") == 1)
	{
  	  my $s_address = $self->get("address");
      my $s_port    = $self->get("port");
      my $dest      = sockaddr_in(27510, inet_aton($s_address));
      $msg = chr(0xff).chr(0xff)."R".chr(0xff).$s_address.chr(0xff).$s_port.chr(0xff).$address.chr(0xff).$port.chr(0xff).$msg.chr(0xff);
 	  $bytes = send($::s_socket, $msg, 0, $dest);
	  &::printEvent("SENDPROXY", "Sent $bytes bytes to proxy server at '$s_address:$s_port'", 1);
    }	
}


sub clear_winner
{
  my ($self) = @_;
  &::printNotice("clear_winner");
  @{$self->{winner}} = ();
}

sub add_round_winner
{
  my ($self, $team) = @_;
  
  # update timestamp from all players. this is much more accurate
  $self->update_players();

  &::printNotice("add_round_winner");
  $self->{winner}[($self->{map_rounds} % 7)] = $team;
  $self->increment("ba_map_rounds");
  $self->increment("map_rounds");
  $self->increment("rounds");
  $self->increment("total_rounds");
  
  $self->{ba_ct_wins} = 0;
  $self->{ba_ts_wins} = 0;
  
  for (@{$self->{winner}}) 
  {
    if ($_ eq "ct")
    {
      $self->increment("ba_ct_wins");
    } elsif ($_ eq "ts")
    {
      $self->increment("ba_ts_wins");
    }
  }
}

sub switch_player
{
  my ($self, $steamid) = @_;
  $self->dorcon("ma_swapteam \"$steamid\"");
  if ($self->get("player_events") == 1)  
  {
    if ($self->get("player_command_steamid") == 1) {
      $cmd_str = $self->get("player_command")." \"$steamid\" You were switched to balance teams";
    } else {
      $cmd_str = $self->get("player_command")." You were switched to balance teams";
    }
	$self->dorcon($cmd_str);
  }	
}


sub analyze_teams
{
  my ($self) = @_;
  
  if ($self->{numplayers} < $self->{minplayers}) 
  {
    &::printNotice("(IGNORED) NOTMINPLAYERS: analyze_teams");
  } 
  elsif ($self->{ba_enabled} > 0)
  {
    &::printNotice("analyze_teams");
    my $ts_skill     = 0;
    my $ts_avg_skill = 0;
    my $ts_count     = 0;
    my $ts_wins      = $self->{ba_ts_wins};
    my $ts_kills     = 0;
    my $ts_deaths    = 0;
    my $ts_diff      = 0;
    my @ts_players   = ();

    my $ct_skill     = 0;
    my $ct_avg_skill = 0;
    my $ct_count     = 0;
    my $ct_wins      = $self->{ba_ct_wins};
    my $ct_kills     = 0;
    my $ct_deaths    = 0;
    my $ct_diff      = 0;
    my @ct_players   = ();
    
    my $server_id   = $self->{id};
    while ( my($pl, $player) = each(::g_players) )
    {	
      my @Player      = ( $player->{server_id},  $player->{name},       $player->{plain_uniqueid}, $player->{skill},   
                          $player->{team},       $player->{map_kills},  $player->{map_deaths},     ($player->{map_kills}-$player->{map_deaths}),
                          $player->{is_dead});
    
      if ($Player[0] == $server_id)
      {
        if ($Player[4] eq "TERRORIST")
        {
          push(@{$ts_players[$ts_count]}, @Player);
          $ts_skill   += $Player[3]; 
          $ts_count   += 1;
          $ts_kills   += $Player[5];
          $ts_deaths  += $Player[6];
        } elsif ($Player[4] eq "CT")
        {
          push(@{$ct_players[$ct_count]}, @Player);
          $ct_skill   += $Player[3]; 
          $ct_count   += 1;
          $ct_kills   += $Player[5]; 
          $ct_deaths  += $Player[6]; 

        }
      }
    }
    
    @ct_players = sort { $b->[7] <=> $a->[7]} @ct_players;
    @ts_players = sort { $b->[7] <=> $a->[7]} @ts_players;
    
    &::printEvent("TEAM", "Checking Teams", 1);
    $admin_msg = "AUTO-TEAM BALANCER: CT ($ct_count) $ct_kills:$ct_deaths  [$ct_wins - $ts_wins] $ts_kills:$ts_deaths ($ts_count) TS";
    if ($self->get("player_events") == 1)  
    {
      $cmd_str = $self->get("player_admin_command")." $admin_msg";
 	  $self->dorcon($cmd_str);
    }	
    $rcon_msg ="AUTO-TEAM BALANCER - CHECKING TEAMS";
    if ($self->get("broadcasting_events") == 1)
	{
       $self->dorcon($self->get("broadcasting_command")." $rcon_msg");
    }   

    if ($self->{ba_map_rounds} >= 2)    # need all players for numerical balacing, at least 2 for getting all players
    {
      my $action_done = 0;
      if ($self->{ba_last_swap} > 0)
      {
        $self->{ba_last_swap}--;
      }
      
      if ($ct_count + 1 < $ts_count)         # ct need players
      {
        $needed_players = floor( ($ts_count - $ct_count) / 2);
        if ($ct_wins < 2)
        {
          @ts_players = sort { $b->[7] <=> $a->[7]} @ts_players;
        } 
        else
        {
          @ts_players = sort { $a->[7] <=> $b->[7]} @ts_players;
        }  
        foreach my $entry (@ts_players) 
        {
          if ($needed_players > 0) # how many we need to make teams even (only numerical)
          {
            if (@{$entry}[8] == 1)  # only dead players!!
            {
              $self->switch_player(@{$entry}[2]); 
              $action_done++;
              $needed_players--;
            }
          }  
        }
      } 
      elsif  ($ts_count + 1 < $ct_count)  # ts need players
      {
        $needed_players = floor( ($ct_count - $ts_count) / 2);
        if ($ts_wins < 2)
        {
          @ct_players = sort { $b->[7] <=> $a->[7]} @ct_players;  # best player
        } 
        else
        {
          @ct_players = sort { $a->[7] <=> $b->[7]} @ct_players;  # worst player
        }  
        foreach my $entry (@ct_players) 
        {
          if ($needed_players > 0) # how many we need to make teams even (only numerical)
          {
            if (@{$entry}[8] == 1)  # only dead players!!
            {
              $self->switch_player(@{$entry}[2]); 
              $action_done++;
              $needed_players--;
            }
          }  
        }
      }
      
      if (($action_done == 0) && ($self->{ba_last_swap} == 0) && ($self->{ba_map_rounds} >= 7) && ($self->{ba_player_switch} == 0)) # frags balancing (last swap 3 rounds before)
      {
         if ($ct_wins < 2)
         {
           if ($ct_count < $ts_count)     # one player less we dont need swap just bring one over
           {
             my $ts_found = 0;
             @ts_players = sort { $b->[7] <=> $a->[7]} @ts_players;  # best player
             foreach my $entry (@ts_players) 
             {
               if ($ts_found == 0)
               {
                 if (@{$entry}[8] == 1)  # only dead players!!
                 {
                   $self->{ba_last_swap} = 3;
                   $self->switch_player(@{$entry}[2]); 
                   $ts_found++;
                 }
               }  
             }  
           }
           else                  # need to swap to players
           {
             my $ts_steamid = 0;
             my $ts_name    = "";
             my $ts_kills   = 0;
             my $ts_deaths  = 0;
             my $ct_steamid = 0;
             my $ct_name    = "";
             my $ct_kills   = 0;
             my $ct_deaths  = 0;
             
             
             my $ts_found = 0;
             @ts_players = sort { $b->[7] <=> $a->[7]} @ts_players;  # best player
             foreach my $entry (@ts_players) 
             {
               if ($ts_found == 0)
               {
                 if (@{$entry}[8] == 1)  # only dead players!!
                 {
                   $ts_steamid = @{$entry}[2];
                   $ts_name    = @{$entry}[1];
                   $ts_kills   = @{$entry}[5];
                   $ts_deaths   = @{$entry}[6];
                   $ts_found++;
                 }
               }  
             }  

             my $ct_found = 0;
             @ct_players = sort { $a->[7] <=> $b->[7]} @ct_players;  # worst player
             foreach my $entry (@ct_players) 
             {
               if ($ct_found == 0)
               {
                 if (@{$entry}[8] == 1)  # only dead players!!
                 {
                   $ct_steamid = @{$entry}[2];
                   $ct_name    = @{$entry}[1];
                   $ct_kills   = @{$entry}[5];
                   $ct_deaths  = @{$entry}[6];
                   $ct_found++;
                 }
               }  
             }  
             if (($ts_found>0) && ($ct_found>0))
             {
               $self->{ba_last_swap} = 3;
               $self->switch_player($ts_steamid); 
               $self->switch_player($ct_steamid); 
             }
           }
         }
         elsif ($ts_wins < 2)
         {
           if ($ts_count < $ct_count)     # one player less we dont need swap just bring one over
           {
             my $ct_found = 0;
             @ct_players = sort { $b->[7] <=> $a->[7]} @ct_players;  # best player
             foreach my $entry (@ct_players) 
             {
               if ($ct_found == 0)
               {
                 if (@{$entry}[8] == 1)  # only dead players!!
                 {
                   $self->{ba_last_swap} = 3;
                   $self->switch_player(@{$entry}[2]); 
                   $ct_found++;
                 }
               }  
             }  
           }
           else                  # need to swap to players
           {
             my $ts_steamid = 0;
             my $ts_name    = "";
             my $ct_steamid = 0;
             my $ct_name    = "";
             
             my $ct_found = 0;
             @ct_players = sort { $b->[7] <=> $a->[7]} @ct_players;  # best player
             foreach my $entry (@ct_players) 
             {
               if ($ct_found == 0)
               {
                 if (@{$entry}[8] == 1)  # only dead players!!
                 {
                   $ct_steamid = @{$entry}[2];
                   $ct_name    = @{$entry}[1];
                   $ct_found++;
                 }
               }  
             }  

             my $ts_found = 0;
             @ts_players = sort { $a->[7] <=> $b->[7]} @ts_players;  # worst player
             foreach my $entry (@ts_players) 
             {
               if ($ts_found == 0)
               {
                 if (@{$entry}[8] == 1)  # only dead players!!
                 {
                   $ts_steamid = @{$entry}[2];
                   $ts_name    = @{$entry}[1];
                   $ts_found++;
                 }
               }  
             }  
             if (($ts_found>0) && ($ct_found>0))
             {
               $self->{ba_last_swap} = 3;
               $self->switch_player($ts_steamid); 
               $self->switch_player($ct_steamid); 
             }
           }
         }
      }
    } # end if rounds > 1  
  }
}

#
# Update player information in database
#

sub updateDB
{
	my ($self) = @_;
	
	my $serverid      = $self->get("id");
	my $rounds        = $self->get("rounds");
	my $kills         = $self->get("kills");
	my $suicides      = $self->get("suicides");
	my $headshots     = $self->get("headshots");
	my $ct_shots      = $self->get("ct_shots");
	my $ct_hits       = $self->get("ct_hits");
	my $ts_shots      = $self->get("ts_shots");
	my $ts_hits       = $self->get("ts_hits");
	my $bombs_planted = $self->get("bombs_planted");
	my $bombs_defused = $self->get("bombs_defused");
	my $ct_wins       = $self->get("ct_wins");
	my $ts_wins       = $self->get("ts_wins");
	my $act_players   = $self->get("numplayers");
	my $act_map       = $self->get("map");
	my $map_rounds    = $self->get("map_rounds");
	my $map_ct_wins   = $self->get("map_ct_wins");
	my $map_ts_wins   = $self->get("map_ts_wins");
	my $map_started   = $self->get("map_started");
	my $map_changes   = $self->get("map_changes");
	my $map_ct_shots  = $self->get("map_ct_shots");
	my $map_ct_hits   = $self->get("map_ct_hits");
	my $map_ts_shots  = $self->get("map_ts_shots");
	my $map_ts_hits   = $self->get("map_ts_hits");
	

    if ($self->{total_kills} == 0)
    {
   	  my $result = &::doQuery("
		  SELECT kills, headshots, suicides, rounds, ct_shots+ts_shots as shots, ct_hits+ts_hits as hits
		  FROM hlstats_Servers
 	      WHERE	serverId='$serverid'
	  ");
  	  ($self->{total_kills}, $self->{total_headshots}, $self->{total_suicides},$self->{total_rounds},$self->{total_shots},$self->{total_hits}) = $result->fetchrow_array();
	}   

	my $result = &::doQuery("
		SELECT count(*) as players
		FROM hlstats_Players
		WHERE game='".$self->get("game")."'
	");
	$self->{players} = $result->fetchrow_array();
	my $players = $self->get("players");
	
	
	# Update player details
	my $query = "
		UPDATE
			hlstats_Servers
		SET  
		    rounds=rounds + $rounds,
			kills=kills + $kills,
			suicides=suicides + $suicides,
			headshots=headshots + $headshots,
			bombs_planted=bombs_planted + $bombs_planted,
			bombs_defused=bombs_defused + $bombs_defused,
			players=$players,
			ct_wins=ct_wins + $ct_wins,
			ts_wins=ts_wins + $ts_wins,
			act_players=$act_players,
			act_map='$act_map',
			map_rounds=$map_rounds,
			map_ct_wins=$map_ct_wins,
			map_ts_wins=$map_ts_wins,
			map_started=$map_started,
			map_changes=map_changes+$map_changes,
			ct_shots=ct_shots + $ct_shots,
			ct_hits=ct_hits + $ct_hits,
			ts_shots=ts_shots + $ts_shots,
			ts_hits=ts_hits + $ts_hits,
			map_ct_shots=$map_ct_shots,
			map_ct_hits=$map_ct_hits,
			map_ts_shots=$map_ts_shots,
			map_ts_hits=$map_ts_hits
		WHERE
			serverId='$serverid'
	";
	&::doQuery($query);

	$self->set("rounds", 0);
	$self->set("kills", 0);
	$self->set("suicides", 0);
	$self->set("headshots", 0);
	$self->set("bombs_planted", 0);
	$self->set("bombs_defused", 0);
	$self->set("ct_wins", 0);
	$self->set("ts_wins", 0);
	$self->set("ct_shots", 0);
	$self->set("ct_hits", 0);
	$self->set("ts_shots", 0);
	$self->set("ts_hits", 0);
	$self->set("map_changes", 0);
}

1;